from typing import Optional

from pydantic import BaseSettings, Extra, SecretStr


class DatabaseConfig(BaseSettings):
    """Config object for connecting to databases.

    This is a generic config object that can be subclassed to create database specific config and
    extended to include specific functionality or defaults relevant to that database technology.

    Notes:
        You probably want to import and use database specific configs such as MySQLConfig and
        PostgresConfig. This is simply a parent class with common config.

    Attributes:
        host: The host name to connect to
        port: The port to use for connections
        user: The username to authenticate with
        password: The password to authenticate with

    """

    host: SecretStr
    port: int
    user: SecretStr
    password: SecretStr

    class Config:
        extra = Extra.ignore


class MySQLConfig(DatabaseConfig):
    """

    Attributes:
        host: The host name to connect to
        port: The port to use for connections
        user: The username to authenticate with
        password: The password to authenticate with
        db: Optional attribute for the database to connect to
    """

    port: int = 3306
    database: Optional[str]


class PostgresConfig(DatabaseConfig):
    port: int = 5432
    database: str



from typing import Dict, Generator, List, Tuple, Union

import pg8000.dbapi as postgresql

from yessql.config import PostgresConfig
from yessql.utils import PendingConnection, PendingConnectionError


class ContextCursor(postgresql.Cursor):
    """**ContextCursor**

    For some reason cursors always require you to define an explicit connect and close. This turns
    these operations into a context manager to make it safer and ensure the cursor is closed after
    use.
    """

    def __enter__(self):
        if isinstance(self.connection, PendingConnection):
            raise PendingConnectionError(
                'Connection to database has not been made. '
                'Did you forget to call the setup_connection() method?'
            )
        else:
            return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class Postgres:
    """**Blocking Postgres Client**

    Synchronous Postgres client for interacting with Postgres databases. For Asynchronous
    Postgres client see yessql.AioPostgres
    """

    def __init__(self, config: PostgresConfig):
        """
        Args:
            config: A PostgresConfig object for connecting to the database
        """
        self.config = config
        self.connection: Union[postgresql.Connection, PendingConnection] = PendingConnection()

    def setup_connection(self) -> None:
        """
        Make a call to the database and attempt to set up a connection. This can either be called
        explicitly or will be called as part of a context statement (I.e. using `with`)

        Returns:
            None
        """
        self.connection = postgresql.connect(
            host=self.config.host.get_secret_value(),
            port=self.config.port,
            user=self.config.user.get_secret_value(),
            password=self.config.password.get_secret_value(),
            database=self.config.database,
        )

    def close_connection(self) -> None:
        """
        Close the connection to the database. Can be used explicitly or will be called as you exit
        out of a context managed statement.

        Returns:
            None
        """
        self.connection.close()

    def __enter__(self):
        self.setup_connection()
        return self

    def __exit__(self, *args):
        self.close_connection()

    def write(self, stmt: str, rows: List[Tuple]) -> int:
        """
        Write some data to the database by passing an insert statement and a list of tuples for the
        rows
        Args:
            stmt: The insert statement to run. `%s` placeholders are used for indicating params
            rows: A list of tuples containing the data to pass to the above query.

        Returns:
            Row count as an integer

        """
        with ContextCursor(connection=self.connection) as cursor:
            cursor.executemany(stmt, rows)
            self.connection.commit()
            return cursor.rowcount

    def commit(self, stmt: str) -> int:
        """
        Although mostly the same as the write method - commit is useful for identifying when you
        are making changes to the database and not pass params. Use this for any CREATE, DROP etc.
        statements where you are managing the database to indicate to other developers where you are
        changing the database VS writing data to it
        Args:
            stmt: The statement to run

        Returns:
            Row count

        """
        with ContextCursor(connection=self.connection) as cursor:
            cursor.execute(stmt)
            self.connection.commit()
            return cursor.rowcount

    def read(self, query: str, params: Tuple = None) -> Generator:
        """
        Read data from the database using the given query and params. This is a generator meaning
        you can iterate through the rows without loading them all into memory.
        Args:
            query: The query to run
            params: Any params to be substituted for `%s` strings in above query

        Returns:
            A generator

        """
        with ContextCursor(self.connection) as cursor:
            if params is not None:
                cursor.execute(query, params)
            else:
                cursor.execute(query)

            keys = [k[0] for k in cursor.description]
            for row in cursor:
                yield dict(zip(keys, row))

    def read_all(self, query: str, params: Tuple = None) -> List[Dict]:
        """
        If you want to return all rows from the query without worrying about memory management
        then this method is useful. It will return a list of dictionaries containing the results
        of the query
        Args:
            query: The query to run
            params: Any params to be substituted for `%s` strings in above query

        Returns:
            A list of dicts with the data from the query

        """
        return [row for row in self.read(query, params)]



from typing import AsyncGenerator, Dict, List, Tuple, Type, Union

from asyncpg import Pool, Record, create_pool
from pydantic import BaseModel

from yessql.aiopostgres.params import NamedParams
from yessql.config import PostgresConfig
from yessql.utils import PendingConnection


class AioPostgres:
    def __init__(
        self, config: PostgresConfig, timeout: int = None, min_size: int = 1, max_size: int = 10
    ):
        """
        AioPostgres is an async postgres client that allows you to set up a connection pool for
        Postgres and read and write data asynchronously. Contains an async context manager for easy
        setup and closing of the connections that you open.
        Args:
            config: a DatabaseConfig object that contains all the connection details for postgres
            timeout: max time before a query is cancelled
            min_size: The minimum # of connections that will be reserved for this client
            max_size: The maximum # of connections that will be reserved for this client
        """
        self.pool: Union[PendingConnection, Pool] = PendingConnection()
        self.config = config
        self.timeout = timeout
        self.min_size = min_size
        self.max_size = max_size

    @property
    def closed(self) -> bool:
        return self.pool._closed  # noqa

    async def setup_pool(self) -> None:
        self.pool = await create_pool(
            host=self.config.host.get_secret_value(),
            port=self.config.port,
            user=self.config.user.get_secret_value(),
            password=self.config.password.get_secret_value(),
            database=self.config.database,
            command_timeout=self.timeout,
            min_size=self.min_size,
            max_size=self.max_size,
        )

    async def __aenter__(self):
        await self.setup_pool()
        return self

    async def read(
        self, query: str, params: Dict = None, model: Type[BaseModel] = None
    ) -> AsyncGenerator:
        """
        Read results from postgres and return an AsyncGenerator. This allows you to read large
        amounts of data without having to store them in memory.
        Args:
            query: The query you want to return data for
            params: Any params you need to pass to the query
            model: An optional pydantic.BaseModel that each row will be parsed to

        Returns:
            An AsyncGenerator
        """
        _params = NamedParams(**params) if params is not None else None
        _query = query.format_map(_params)

        async with self.pool.acquire() as conn:  # type: ignore
            async with conn.transaction():
                cur = conn.cursor(_query, *_params.as_tuple) if _params else conn.cursor(_query)
                async for row in cur:
                    if model:
                        yield model(**row)
                    else:
                        yield row

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close_pool()

    async def read_all(
        self, query: str, params: Dict = None, model: Type[BaseModel] = None
    ) -> Union[List[Record], Type[BaseModel]]:
        """
        In some cases you might want to just return the data without dealing with iteration you can
        use this. We'll return all the records in a list. Be careful using this for large datasets
        as it will try and load everything in memory.
        Args:
            query: The query you want to return data for
            params: Any params you need to pass to the query
            model:

        Returns:
            A List of Records
        """
        rows = []
        async for row in self.read(query=query, params=params, model=model):
            rows.append(row)
        return rows

    async def write(self, stmt: str, params: Tuple) -> None:
        """
        Write data to a table with the given statement and data
        Args:
            stmt: The Insert statement you want to run
            params: The data to pass as params

        Returns:
            None
        """
        async with self.pool.acquire() as conn:  # type: ignore
            async with conn.transaction():
                await conn.executemany(stmt, params)

    async def commit(self, stmt: str) -> None:
        """
        Run a command against the database. This is useful for statements where you need to change
        the database in some way E.g. ALTER, CREATE, DROP statements etc.
        Args:
            stmt: The statement to run
        """
        async with self.pool.acquire() as conn:  # type: ignore
            await conn.execute(stmt)

    async def close_pool(self) -> None:
        await self.pool.close()  # type: ignore

    def writer(self, stmt: str):
        """return a writer for the given statement.

        Basically just curry's the write method into a coroutine that accepts a batch of parameters
        and writes using the given statement. Obviously this is useful if you are batching inserts.

        Args:
            stmt: The Insert statement you want to run
        """

        async def writer(batch):
            await self.write(stmt, batch)

        return writer



__version__ = '0.1.3'
__author__ = 'Mitchell Lisle'
__email__ = 'm.lisle90@gmail.com'


from yessql.aiomysql import AioMySQL
from yessql.aiopostgres.client import AioPostgres
from yessql.aiopostgres.params import NamedParams
from yessql.config import DatabaseConfig, MySQLConfig, PostgresConfig
from yessql.logger import logger
from yessql.postgres import ContextCursor, Postgres
from yessql.utils import PendingConnection, PendingConnectionError



from typing import AsyncGenerator, Dict, List, Tuple, Union

import aiomysql as mysql

from yessql.config import MySQLConfig
from yessql.utils import PendingConnection


class AioMySQL:
    def __init__(
        self,
        config: MySQLConfig,
        cursor_class: mysql.Cursor = mysql.SSDictCursor,
        min_size: int = 1,
        max_size: int = 10,
    ):
        """
        AioMySQL is an async postgres client that allows you to set up a connection pool for
        MySQL and read and write data asynchronously. Contains an async context manager for easy
        setup and closing of the connections that you open.
        Args:
            config: a DatabaseConfig object that contains all the connection details for postgres
            cursor_class: The cursor class to use
            min_size: The minimum # of connections that will be reserved for this client
            max_size: The maximum # of connections that will be reserved for this client
        """
        self.pool: Union[mysql.Pool, PendingConnection] = PendingConnection()
        self.config: MySQLConfig = config
        self.cursor_class: mysql.Cursor = cursor_class
        self.min_size: int = min_size
        self.max_size: int = max_size

    async def setup_pool(self):
        self.pool = await mysql.create_pool(
            host=self.config.host.get_secret_value(),
            user=self.config.user.get_secret_value(),
            password=self.config.password.get_secret_value(),
            db=self.config.database,
            port=self.config.port,
            minsize=self.min_size,
            maxsize=self.max_size,
        )

    async def __aenter__(self):
        await self.setup_pool()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close_pool()

    async def read(self, query: str, params: Tuple = None) -> AsyncGenerator:
        async with self.pool.acquire() as conn:  # type: ignore
            async with conn.cursor(self.cursor_class) as cur:
                await cur.execute(query, params)
                async for row in cur:
                    yield row

    async def read_all(self, query: str, params: Tuple = None) -> List[Dict]:
        rows = []
        async for row in self.read(query, params):
            rows.append(row)
        return rows

    async def write(self, stmt: str, params: Union[Tuple, str, int]) -> None:
        async with self.pool.acquire() as conn:  # type: ignore
            async with conn.cursor() as cur:
                await cur.executemany(stmt, params)
            await conn.commit()

    async def commit(self, stmt: str):
        async with self.pool.acquire() as conn:  # type: ignore
            async with conn.cursor() as cur:
                await cur.execute(stmt)
            await conn.commit()

    async def close_pool(self):
        self.pool.close()
        await self.pool.wait_closed()



from typing import Tuple


class NamedParams(dict):
    @property
    def as_tuple(self) -> Tuple:
        return tuple(self.values())

    def __getitem__(self, key):
        return f'${list(self.keys()).index(key) + 1}'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
from setuptools import find_packages, setup

with open('requirements.txt') as f:
    requirements = f.read().splitlines()


with open('requirements/test.txt') as f:
    test_requirements = f.read().splitlines()


setup(
    author='Mitchell Lisle',
    author_email='m.lisle90@gmail.com',
    description='An easy to use python SQL database interface for Postgres and MySQL',
    install_requires=requirements,
    include_package_data=True,
    keywords='yessql',
    name='yessql',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/mitchelllisle/yessql',
    version='0.1.3',
    zip_safe=False,
)



class PendingConnectionError(ValueError):
    """Error for when you attempt to use connection before calling setup_connection"""


class PendingConnection:
    """Pending Connection Class
    This class is used to differentiate between having made an actual connection to the database
    vs only having instantiated the class. This is important so that we can catch any confusing
    errors and instead return more helpful messages explaining that a connection has not been made
    yet.
    """

    def __enter__(self):
        raise PendingConnectionError(
            'You are trying to use a PendingConnection object. '
            'Did you forget to run setup_connection()?'
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def close(self):
        pass

    def commit(self):
        pass
